#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['ui3', 'ui3.anchor']

package_data = \
{'': ['*']}

setup(name='ui3',
      version='2021.1.10',
      description='UI and other utils for Pythonista iOS app',
      author='Mikael Honkala',
      author_email='mikael.honkala@gmail.com',
      url='https://github.com/mikaelho/uiutils',
      packages=packages,
      package_data=package_data,
     )
